#! /usr/bin/python3

import logging
import datetime
from datetime import timedelta
import sys
import os
from os import getenv
import time
from time import sleep
from shutil import copyfileobj
import urllib.request
import urllib.error
import pathlib
import contextlib
import fcntl

from csppfetch.exclusivelockfile import AtomicCreateIfMissing
import csppfetch.daterange
import csppfetch.timelimit

ENV_URL_SUFFIX = "URL"
ENV_TIMEOUT_SUFFIX = "TIMEOUT"
ENV_RETRIES_SUFFIX = "RETRIES"
ENV_RETRY_WAIT_SUFFIX = "RETRY_WAIT"

# Dates this far into the future are highly dubious and will trigger errors.
IMPLAUSIBLY_DISTANT_FUTURE = datetime.datetime.now() + timedelta(days=365*10)

MIN_PYTHON="3.6"
MIN_PYTHON_TUPLE=tuple([int(x) for x in MIN_PYTHON.split(".")])

if sys.version_info < MIN_PYTHON_TUPLE:
    f"If this fails, you're using too old of a Python, you need at least the value of MIN_PYTHON in this file"
    raise Exception(f"Requires Python {MIN_PYTHON} or newer.")








class FileSet:
    """ Represents a logical group of files to be downloaded

    The set implies all-or-nothing; if one or more files is
    unavailable, the FileSet as a whole is not useful for the
    current purpose.
    """
    def __init__(self, urls_to_files, expected = False, description=None):
        """ Create a FileSet

        This is primarily intended for use of Downloader objects, which will
        construct and return a list of FileSets with Downloader.get_filesets_for_time()

        urls_to_files - Dict-like map of RELATIVE URLs to RELATIVE local
        file paths

        expected - If True, one or more unavailable files should log a
        logging.WARNING. If False only a logging.INFO or logging.DEBUG should
        be logged.

        description - Brief human readable description for logging

        >>> a = FileSet( {'path/to/fileA': 'fileA', 'some/other/file': 'fileB'}, True, 'Example Data')
        >>> a.description
        'Example Data'
        >>> a.expected
        True
        >>> a.urls_to_files
        {'path/to/fileA': 'fileA', 'some/other/file': 'fileB'}
        """
        self.urls_to_files = urls_to_files
        self.expected = expected
        self.description = description

    def multiline_str(self):
        r = [f"{self.description} (expected? {self.expected}) containing"]
        for url in sorted(self.urls_to_files):
            r += [ f"  {url}: {self.urls_to_files[url]}" ]
        return r

    def __repr__(self):
        maps = [ f"{url}: {self.urls_to_files[url]}" for url in sorted(self.urls_to_files)]
        maps = ", ".join(maps)
        r = f"<{self.description} (expected? {self.expected}): {{{maps}}}>"
        return r

    def __eq__(self, other):
        return (
                (self.description == other.description) and
                (self.expected == other.expected) and
                (self.urls_to_files == other.urls_to_files)
                )



class DownloadStatistics: 
    def __init__(self):
        self.downloaded_files = 0
        self.downloaded_size = 0
        self.cache_hits = 0
        self.cache_size = 0
        self.deleted_files = 0
        self.deleted_size = 0

    def __str__(self):
        return (f"Downloaded {self.downloaded_files} files "
            f"totalling {self.downloaded_size} bytes. "
            f"Pulled {self.cache_hits} files from cache, "
            f"saving {self.cache_size} bytes of download. "
            f"Deleted {self.deleted_files} files, "
            f"freeing {self.deleted_size} bytes of space. "
            )


def download_successful(state, dst, size, download_stats):
    """ Do bookkeeping when a desire file has been downloaded
    """
    logging.debug(f"{dst} downloaded ({state}).")
    if download_stats is not None:
        download_stats.downloaded_files += 1
        download_stats.downloaded_size += size


def download_already_present(state, dst, download_stats, i_own_file):
    """ Do bookkeeping when a desire filed already exists

    Intended for use by download_all

    state - Human readable description for logging

    dst - File that exists

    download_stats - Optional DownloadStatistics object; will be updated

    i_own_file - If True, we should be able to modify timestamp on dst and
    failing to do so will log an error and raise an exception. If False, it's
    okay if modifying the timestamp fails. 

    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile() as f:
    ...     download_already_present("Example Download", f.name, None, True)
    """
    logging.debug("{0} already exists ({1}). Skipping.".format(dst, state))
    if download_stats is not None:
        download_stats.cache_hits += 1
        download_stats.cache_size += os.path.getsize(dst)
    try:
        pathlib.Path(dst).touch()
    except PermissionError as e:
        msg = 'Unable to update timestamp on "{0}": {1}'.format(dst, e)
        if i_own_file:
            logging.error(msg)
            logging.error("This is a problem as I expected to be able to manage the cache. If I should treat the cache as read-only, use '--local-only' (if you ran aitf-mirror) or '--cache-only' (if you ran aitf).")
            raise Exception(msg)
        else:
            logging.info(msg)
            logging.info("Assuming I'm not allowed to update the cache and that this is okay.")
    except Exception as e:
        msg = 'Unable to update timestamp on "{0}": {1}'.format(dst, e)
        raise Exception(msg)

HUGE_TIMEOUT=60*60*24*30 # 30 days in seconds
def download_to_file_no_lock(url, dst, fileobj, timeout=HUGE_TIMEOUT):
    """ Download file from url to fileobj

    dst is used for logging and need not correspond to where
    fileobj will put the data.

    Intended for use by download_all.

    >>> import tempfile,os
    >>> with tempfile.TemporaryDirectory() as dir:
    ...     dst = dir+"/test"
    ...     with open(dst,"wb") as f:
    ...         download_to_file_no_lock("http://www.google.com/", dst, f)
    ...     os.path.getsize(dst) > 0
    True
    """

    more = ""
    if hasattr(fileobj, 'name'):
        more = f"(using temporary file {fileobj.name})"
    logging.debug(f"Downloading {url} to {dst} {more}")
    with timelimit.TimeLimit(timeout):
        with urllib.request.urlopen(url) as req:
            copyfileobj(req,fileobj)


def download_all(url_to_dst, download_stats, do_download, timeout):
    """ Download a Dict of URL keys to their matching local values

    Files which are already present are _skipped_. The remaining files
    are downloaded atomically; all or nothing.
    
    You probably don't want this, but instead want something from 
    dataset, maybe dataset.get_first_set or dataset.mirror
    
    url_to_dst is a Dict where the key is a URL to download and value is
    is the local path to write it to.  If the local path is already present,
    that download will be skipped.

    download_status is None or a DownloadStatistics object.

    do_download is True if we should download files not present locally.
    If False, if any files are not already locally present, an Exception
    is raised.

    timeout is maximum seconds a given download (of a single file) is allowed
    to take before it's abndoned and considered a failure.

    Raises an Exception if at least one desired file is not available.

    >>> import tempfile, os
    >>> with tempfile.TemporaryDirectory() as dirpath:
    ...     download_all({
    ...         "http://www.google.com/": dirpath+"/a",
    ...         "http://www.microsoft.com/": dirpath+"/b",
    ...         }, None, True, 60*5)
    ...     [os.path.exists(dirpath+X) for X in ("/a","/b")]
    [True, True]
    """

    # Prune files we already have.
    # Redundant with a later check, but this minimizes lock contention
    todo = {}
    for src,dst in url_to_dst.items():
        if os.path.exists(dst):
            download_already_present("download_all pre-lock",
                    dst, download_stats, do_download)
        else:
            todo[src] = dst
    url_to_dst = todo

    with contextlib.ExitStack() as lock_stack:
        todo = {}
        # Lock files we're about to download into
        for url,dst in url_to_dst.items():
            atomic = AtomicCreateIfMissing(dst)
            file = atomic.__enter__()
            if atomic is None:
                atomic.__exit__(None,None,None)
                download_already_present("download_all post-lock",
                        dst, download_stats, do_download)
                continue

            lock_stack.push(atomic)
            todo[url] = (dst,file,atomic)

        if (len(todo) > 0) and not do_download:
            raise Exception("One or more files do not exist and downloading is disabled. "+str(todo))

        # After all that build up, I'm a little let down by the punchline
        try:
            for url,dst in todo.items():
                (path,file,atomic) = dst
                download_to_file_no_lock(url, path, file, timeout)
                download_successful("download_all success", path, atomic.final_size(), download_stats)

        except:
            # Files present at start should not be in todo, so this is just
            # stuff we might have downloaded. Wipe it out, as it's all or
            # nothing
            logging.debug("Failed to acquire all desired files in FileSet")
            if(len(todo.values())):
                logging.debug("Attempting to remove up files downloaded as part of this failed FileSet")
            for path,file,atomic in todo.values():
                if atomic.abandon():
                    logging.debug(f"    {dst} removed")
            raise



class Downloader:
    """ Knowledge of how to download a given data source's files
    """

    def __init__(self,
            name,
            package_env_id,
            url_base,
            url_relative,
            local,
            period,
            epoch_start = datetime.datetime(2000,1,1,0,0,0),
            priorities = None,
            oldest_usable = timedelta(days=7),
            oldest_mirror = timedelta(days=7),
            expected_newest = None,
            ):
        """
            name   - Brief human readable description, for logging
            package_env_id - environment variable prefix this package uses
                     to override parts of this.
                     For example: "CSPP_GEO_AITF_ANCIL_"
            url_base - Default base URL files are available from.
                     User overridable via environment.
                     Can contain multiple entries, comma separated (so
                     you'll need to escape commas in the URLs!)
                     Should be entirely static, no % or {} formatting.
            url_relative - relative URL to download data from. Can and probably
                     should contain % formatting codes suitable to pass to
                     datetime.datetime.strftime.
                     May contain '{priority}' which will receive an entry 
                     from priorities if available.
            local -  relative path for a recommended filename to write the 
                     downloaded file to. May contain directory entries.
                     Can and probably should contain % formatting codes
                     suitable to pass to datetime.datetime.strftime.
                     May contain '{priority}' which will receive an entry 
                     from priorities if available.
            period - datetime.timedelta describing how often the source
                     data is provided
            epoch_start -  datetime.datetime from which periods are measured.
                     No data should exist prior to this date.
            priorities - If not None, an ordered list of options in preferance
                     order. ex: SST might use ['', '_preliminary'] (as the
                     final data doesn't have any indicator.)
            oldest_usable - datetime.timedelta when searching for suitable data
                     for a given goal datetime, no data more than 72 hours
                     before the goal datetime will be offered.
            oldest_mirror - datetime.timedelta for how far back to maintain a
                     mirror.
            expected_newest - If not None, a datetime.timedelta for how far
                     before "now" we expect data to be available. Data before
                     then should log at logging.WARNING about missing data;
                     after that point is logging.INFO. If None, always us INFO

            This implies that data is potentially available at (epoch_start +
            period*N) for integers N from 0 through infinity

        >>> from datetime import timedelta, datetime
        >>> sst_filename = "avhrr-only-v2.%Y%m%d{priority}.nc"
        >>> sst = Downloader(
        ...     name = "Sea Surface temperature",
        ...     package_env_id = 'CSPP_GEO_AITF_ANCIL_',
        ...     url_base = "https://geodb.ssec.wisc.edu/ancillary/",
        ...     url_relative = "%Y_%m_%d_%j/"+sst_filename,
        ...     local = sst_filename,
        ...     period = timedelta(days=1),
        ...     epoch_start = datetime(2010,1,1,0,0,0),
        ...     # Empty string priority is useful if a file may omit the priority
        ...     priorities = ['', '_preliminary'],
        ...     oldest_usable = timedelta(days=7),
        ...     expected_newest = timedelta(hours=1)
        ...     )
        """
        if oldest_mirror < timedelta(days=0):
            raise ValueError(f"oldest_mirror should be positive; it is {oldest_mirror}")
        if oldest_usable < timedelta(days=0):
            raise ValueError(f"oldest_usable should be positive; it is {oldest_usable}")
        if epoch_start > IMPLAUSIBLY_DISTANT_FUTURE:
            raise ValueError(f"oldest_usable should be positive; it is {oldest_usable}")

        self.name = name
        self.package_env_id = package_env_id
        self.url_base = url_base
        self.url_relative = url_relative
        self.local = local
        self.priorities = priorities
        self.period = period
        self.epoch_start = epoch_start
        self.oldest_usable = oldest_usable
        self.oldest_mirror = oldest_mirror
        self.expected_newest = expected_newest

    def _nearest_preceeding_time(self, time):
        """ Get datetime of data most recent to time

        time should be a datetime.datetime

        returns a datetime.datetime at the same time as or before time
        
        returns None if no valid answer is possible (asking for data before
        the epoch_start)

        >>> import datetime
        >>> d._nearest_preceeding_time(datetime.datetime(2020,6,1,12,0,0))
        datetime.datetime(2020, 6, 1, 3, 0)
        >>> d._nearest_preceeding_time(datetime.datetime(2020,6,1,3,0,0))
        datetime.datetime(2020, 6, 1, 3, 0)

        """
        delta = time - self.epoch_start
        num_periods = int(delta / self.period)
        return self.epoch_start + self.period * num_periods

    def _previous_time(self, time):
        """ One self.period earlier than time

        This does take into account self.epoch_start, so
        the returned time will be in the range
        "time" through "time-period*2" EXCLUSIVE of
        both ends.

        >>> import datetime
        >>> d._previous_time(datetime.datetime(2020,6,1,12,0,0))
        datetime.datetime(2020, 6, 1, 3, 0)
        >>> d._previous_time(datetime.datetime(2020,6,1,3,0,0))
        datetime.datetime(2020, 5, 31, 15, 0)
        """
        t = self._nearest_preceeding_time(time)
        if t != time:
            return t
        return t - self.period

    def _daterange(self, start, end, inclusive = False):
        """ return a daterange from start to end

        Automatically aligns to self.epoch_start

        ONLY SUPPORTS start > end (that is, the start is a later
        datetime than end.

        >>> from datetime import datetime
        >>> start = datetime(2020,6,1, 12,0,0)
        >>> end = datetime(2020,5,30)
        >>> list(d._daterange(start, end))
        [datetime.datetime(2020, 6, 1, 3, 0), datetime.datetime(2020, 5, 31, 15, 0), datetime.datetime(2020, 5, 31, 3, 0), datetime.datetime(2020, 5, 30, 15, 0), datetime.datetime(2020, 5, 30, 3, 0)]
        """
        if start < end: # TODO: fix silly limitation
            raise ValueError("start must be AFTER end. Forward iteration is not yet supported")
        step = -self.period
        start = self._nearest_preceeding_time(start)
        return daterange.daterange(start, end, step, inclusive)

    def _priorities(self, time):
        """ Get potential priorities for time

        Default implementation assumes all priorities are available
        at all times.

        >>> import datetime
        >>> d._priorities(datetime.datetime(2020,1,1))
        ['final', 'preliminary']
        """
        if self.priorities is None:
            return ['']
        return self.priorities

    def _expand(self, path, time, priority):
        return time.strftime(path).format(priority=priority)

    def get_mirror_filesets(self, start = None, end = None):
        """ Return FileSets to create a mirror between start and end, inclusive

        Return a list of lists of FileSets which represent all files that
        should be downloaded to create a mirror between start and end.

        The top level list is one entry per timestep; all entries in this
        list should be mirrored without concern for success or failure of
        other entries.

        The sublist is a list of FlieSets in priority order; each entry
        should be tried in order, stopping once on is downloaded or already avialable.

        end defaults to "now", and start defaults to end minus self.oldest_mirror

        So in pseudocode:

            from datetime import datetime
            mirror = d.get_mirror_filesets(datetime(2020,1,1), datetime(2021,2,1))
            for preference_list in mirror:
                for item in preference_list:
                    if download(item):
                        break

        The standard Downloader.get_mirror_filesets implementation is highly
        likely to return the same file in multiple lists. This is intentional
        and should be harmless.

        >>> import datetime
        >>> timesteps = d.get_mirror_filesets(end=datetime.datetime(2021,7,14,12,0,0))
        >>> for i, timestep in enumerate(timesteps): # doctest: +ELLIPSIS
        ...     print(f"timestep {i}")
        ...     for fileset in timestep: print(fileset.description)
        timestep 0
        Example Downloader for 2021-07-14 03:00:00 at priority "final"
        Example Downloader for 2021-07-14 03:00:00 at priority "preliminary"
        Example Downloader for 2021-07-13 15:00:00 at priority "final"
        Example Downloader for 2021-07-13 15:00:00 at priority "preliminary"
        Example Downloader for 2021-07-13 03:00:00 at priority "final"
        Example Downloader for 2021-07-13 03:00:00 at priority "preliminary"
        timestep 1
        Example Downloader for 2021-07-13 15:00:00 at priority "final"
        ...
        timestep 3
        Example Downloader for 2021-07-12 15:00:00 at priority "final"
        Example Downloader for 2021-07-12 15:00:00 at priority "preliminary"
        Example Downloader for 2021-07-12 03:00:00 at priority "final"
        Example Downloader for 2021-07-12 03:00:00 at priority "preliminary"
        Example Downloader for 2021-07-11 15:00:00 at priority "final"
        Example Downloader for 2021-07-11 15:00:00 at priority "preliminary"
        """

        if self.oldest_mirror < timedelta(days=0):
            raise ValueError(f"oldest_mirror should be positive; it is {self.oldest_mirror}")

        if end is None: end = datetime.datetime.now()
        if start is None: start = end - self.oldest_mirror

        if start > end:
            raise ValueError(f"start ({start}) should be before end ({end})")

        fileset_list_list = []
        for time in self._daterange(end, start, inclusive=True):
            fileset_list = self.get_filesets_for_time(time)
            fileset_list_list.append(fileset_list)
        return fileset_list_list


    def get_filesets_for_time(self, time):
        """ Return list of FileSets suitable for use at time

        If time represents data you want to process, returns a list
        of FileSets that could be used as ancillary data for that time.
        Any entry on the list is individually complete; callers will
        want to try downloading each entry in the list until a full
        entry is available. The list is sorted in preference order, which
        by default is likely preferring newer data to older data and
        then preferring files from earlier in self.priorities to later.

        >>> import datetime
        >>> fs = d.get_filesets_for_time(datetime.datetime(2020,6,1,12,0,0))
        >>> for f in fs: print(f.description)
        Example Downloader for 2020-06-01 03:00:00 at priority "final"
        Example Downloader for 2020-06-01 03:00:00 at priority "preliminary"
        Example Downloader for 2020-05-31 15:00:00 at priority "final"
        Example Downloader for 2020-05-31 15:00:00 at priority "preliminary"


        #>>> a = "\n".join([ f.description for f in fs])
        """
        earliest_time = time - self.oldest_usable

        filesets = []
        for time in self._daterange(time, earliest_time, inclusive=True):
            for priority in self._priorities(time):
                url_relative = self._expand(self.url_relative, time, priority)
                local = self._expand(self.local, time, priority)
                urls_to_files = {url_relative: local}

                expected = False
                if self.expected_newest is not None:
                    expected = time < (datetime.datetime.now() - self.expected_newest )

                priority_desc = f'"{priority}"' if priority != "" else '"" (empty string)'
                description = f"{self.name} for {time} at priority {priority_desc}"

                fs = FileSet(urls_to_files, expected, description)
                filesets.append(fs)

        return filesets


    def get_base_urls(self):
        """ Returns list of base URLs where data might be found

        Draws from self.urlbase and the environment variable
        identifier+"URL".

        Thing that lack ":/" are assumed to be local paths and
        have "file://" prepended.

        >>> d.get_base_urls()
        ['https://geodb.ssec.wisc.edu/ancillary/']

        Assuming Downloader was created with package_env_id="EXAMPLE_":

        >>> import os
        >>> os.environ['EXAMPLE_URL'] = "http://example.com/,ftp://ftp.example.com/pub/"
        >>> d.get_base_urls()
        ['http://example.com/', 'ftp://ftp.example.com/pub/']

        """
        envname = "NO VALID ENVIRONMENT VARIABLE. to allow environment variables to override the defaults, set self.package_env_id to a valid string"
        value = self._getenv(ENV_URL_SUFFIX, self.url_base)
        if value is None or len(value) == 0:
            raise Exception(f"Internal error: No valid URL bases for {self.name}. Initialized without a valid URL base and environment variable {envname} is empty.")
        rawurls = value.split(',')
        checkedurls = []
        for u in rawurls:
            if ':/' not in u:
                logging.warning(f"Assuming that {u} (from {envname}) should be file:///{u}")
                u = "file:///"+u
            checkedurls.append(u)
        return checkedurls

    def _getenv(self, name_suffix, default, datatype=str):
        """ Retrieve download configuration environment variable

        The suffix (ex "URL" or "RETRIES") will be appended to the
        package_env_id (ex "CSPP_GEO_AITF_ANCIL_"). If that environment
        variable exists, and it can be cast into datatype, that casted value
        is returned. Otherwise returns default.

        In these examples, d is a Downloader with packagen_env_id="EXAMPLE_".
        The name name_suffix should come from ENV_*_SUFFIX, defined in the
        package

        >>> d._getenv("FAKE", "default value")
        'default value'
        >>> from os import environ
        >>> environ["EXAMPLE_FAKE"] = "99"
        >>> d._getenv("FAKE", "default value")
        '99'
        >>> d._getenv("FAKE", 101, int)
        99
        >>> environ["EXAMPLE_FAKE"] = "value changed"
        >>> d._getenv("FAKE", "default value", int)
        Traceback (most recent call last):
        TypeError: Internal error: default value default value is of type <class 'str'> which is not a <class 'int'>
        """
        if not isinstance(default, datatype):
            raise TypeError(f"Internal error: default value {default} is of type {type(default)} which is not a {datatype}")

        is_default = True
        value = default
        name = name_suffix
        if self.package_env_id is not None:
            name = self.package_env_id+name_suffix
            raw_value = getenv(name, None)
            if raw_value is not None:
                try:
                    value = datatype(raw_value)
                    is_default = False
                except ValueError as v:
                    logging.error(f"{name} is {raw_value}, which does not appear to be of type {datatype}. Using default of {value} instead.")
        msg = f"{name}={value}"+ " (default value)" if is_default else ""
        logging.info(msg)
        return value


    def get_timeout(self, timeout):
        """ Return timeout in seconds; letting environment override if present """
        return self._getenv(ENV_TIMEOUT_SUFFIX, timeout, int)

    def get_retries(self, retries):
        """ Return retry limit; letting environment override if present """
        return self._getenv(ENV_RETRIES_SUFFIX, retries, int)

    def get_retry_wait(self, retry_wait):
        """ Return seconds delay between retries; letting environment override if present """
        return self._getenv(ENV_RETRY_WAIT_SUFFIX, retry_wait, int)

    def download_first_available(self, filesets, dst, timeout, retries, retry_wait, do_download = True, download_stats = None):
        """ Download first available FileSet from list filesets, writing into dst

        If a desired file is already present, the download of that file is
        considered a success and not re-downloaded.

        timeout - seconds to allow a given download to try before abandoning it
        as failed.

        retries - total number of times to try downloading each FileSet from
        each base URL before considering the entire attempt failed.

        retry_wait - seconds to wait after failing to download any FileSet from
        any base URL.

        do_download - If True, files not already present will be downloaded.
        If False, if the desired files are not already present, it's a failure

        download_stats - Optional DownloadStatistics object to collect download
        statistics

        """
        base_urls = self.get_base_urls()

        problems = []
        def dump_problems(logfunc,problems):
            last_problem = None
            for problem in problems:
                if problem != last_problem:
                    logfunc("  "+problem)
                    last_problem = problem

        for attempt in range(1, retries+1):
            for fileset in filesets:
                description = fileset.description
                expected = fileset.expected
                urls_to_files = fileset.urls_to_files

                for base_url in base_urls:
                    full_urls_to_file = {base_url+suffix: dst+"/"+value for suffix,value in urls_to_files.items()}

                    try:
                        logging.debug(f"\n\nConsidering {full_urls_to_file}\n")
                        download_all(full_urls_to_file, download_stats, do_download, timeout)
                        try:
                            logging.info("Data set successfully acquired. Problems encountered and resolved:")
                            dump_problems(logging.info, problems)
                        except:
                            # Don't cause grief over logging problems here
                            # since any exceptions will retry downloads
                            # when we already have the files!
                            pass
                        return True
                    except Exception as e:
                        problems.append(f"Failed to download file set {description}: {str(e)}")
            if attempt != retries:
                sleep(retry_wait)
        logging.error("Unable to download any file set, reasons include:")
        dump_problems(logging.error, problems)

    def download_for_time(self, time, dst, timeout=30, retries=3, retry_wait=20, do_download = True, download_stats = None):
        """ Download files needed to process data a time, writing into dst

        If a desired file is already present, the download of that file is
        considered a success and not re-downloaded.

        timeout - seconds to allow a given download to try before abandoning it
        as failed.

        retries - total number of times to try downloading each FileSet from
        each base URL before considering the entire attempt failed.

        retry_wait - seconds to wait after failing to download any FileSet from
        any base URL.

        do_download - If True, files not already present will be downloaded.
        If False, if the desired files are not already present, it's a failure

        download_stats - Optional DownloadStatistics object to collect download
        statistics

        timeout, retries, and retry_wait are overriddeen by environment
        variables package_env_id+TIMEOUT/RETRIES/RETRY_WAIT

        """

        filesets = self.get_filesets_for_time(time)
        
        timeout = self.get_timeout(timeout)
        retries = self.get_retries(retries)
        retry_wait = self.get_retry_wait(retry_wait)
        self.download_first_available(filesets, dst, timeout=timeout, retries=retries, retry_wait=retry_wait, do_download = do_download, download_stats = download_stats)


    def mirror(self, dst, start_time=None, end_time=None, timeout=30, download_stats=None):
        """ Mirror all files for processing times beween start_time and end_time

        start_time and end_time are datetime.datetimes.

        end defaults to "now", and start defaults to end minus self.oldest_mirror

        If a desired file is already present, the download of that file is
        considered a success and not re-downloaded.

        timeout - seconds to allow a given download to try before abandoning it
        as failed. Can be overriden by environment variable
        package_env_id+"TIMEOUT"

        download_stats - Optional DownloadStatistics object to collect download
        statistics
        """
        if end_time is not None and start_time is not None:
            time = self._nearest_preceeding_time(end_time)
            if time < start_time:
                logging.warning("Mirror requested for range {start_time} through {end_time}, but the preceeding available time is {time}, which is before {start_time}. Nothing will happen.")

        logging.info("Attempting to mirror data for processing between {start_time} and {end_time}")

        filesets_list_list = self.get_mirror_filesets(start_time, end_time)
        timeout = self.get_timeout(timeout)

        for filesets in filesets_list_list:
            self.download_first_available(filesets, dst, timeout=timeout, retries=1, retry_wait=1, download_stats = download_stats)




    def clean_local_mirror(self, oldest_valid):
        """ Remove files with timestamps older than oldest_valid

        oldest_valid is a datetime.datetime
        """
        TODO



