import os
import fcntl
import logging

class ExclusiveLockFile:
    """ Lock a file by name or handle, returning an open file handle

    Can be used in a "with" block as a context manager.

    >>> from csppfetch.exclusivelockfile import ExclusiveLockFile
    >>> from tempfile import mkdtemp
    >>> tmpdir = mkdtemp()
    >>> filename = tmpdir+"/lock"
    >>>
    >>> # You can't lock a file until it exists
    >>>
    >>> ExclusiveLockFile(filename) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ... 
    FileNotFoundError: [Errno 2] No such file or directory: '.../lock'
    >>>
    >>> # You can lock a file-like object that support f.name and fcntl.lockf(f)
    >>>
    >>> f = open(filename,"w")
    >>> with ExclusiveLockFile(f):
    ...     # Can assume exclusive control of filename
    ...     pass
    >>>
    >>> # You can lock an existing file by name (str, bytes, or Path-like object).
    >>>
    >>> with ExclusiveLockFile(filename) as f:
    ...     # Can assume exclusive control of filename
    ...     # Can use f as a file object
    ...     pass
    >>>
    >>> # Also supports manual use; you MUST call unlock in this case!
    >>>
    >>> l = ExclusiveLockFile(filename)
    >>> l.lock()
    >>> # Can assume exclusive control of filename
    >>> l.unlock()
    >>> 
    >>> # Error cases
    >>> l.unlock()  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    RuntimeError: Attempting to unlock .../lock, but I don't have that lock
    >>> l.lock()
    >>> l.lock()    # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    RuntimeError: Attempting to lock .../lock, but I already have that lock
    >>>
    >>> import os
    >>> os.unlink(filename)
    >>> os.rmdir(tmpdir)
    """

    def __init__(self, filename_or_file):
        """ prepare to lock filename_or_file

        filename_or_file can be str or bytes path to a file,
        a Path-like object, or a file object.

        file-like objects will only work to the extent that 
        fcntl.lockf works on them.

        Extended examples are in help(exclusivelockfile.ExclusiveLockFile)
        """
        try: 
            # Ensure it's Path-like
            os.fspath(filename_or_file)
            self.file = None
            self.filename = filename_or_file
            logging.debug(f"ExclusiveLockFile({filename_or_file}) is path-like")
        except:
            # Not Path-like? I hope it's file-like
            self.file = filename_or_file
            self.filename = None
            logging.debug(f"ExclusiveLockFile({type(filename_or_file)} {filename_or_file.name}) is hopefully file-like")

        if self.file is None:
            # r+ so we don't create the file if it doesn't alreasy exist.
            self.file = open(filename_or_file, "r+b")
        else:
            try:
                self.filename = self.file.name
            except AttributeError:
                raise TypeError(f"filename_or_file passed in does not appear to a a str, bytes, Path-like object, or a file object. It is a {type(filename_or_file)}")

        self.locked = False

    def lock(self):
        if self.locked:
            raise RuntimeError(f"Attempting to lock {self.filename}, but I already have that lock")
        stat_res = os.stat(self.file.fileno())
        logging.debug(f"ExclusiveLockFile.lock({self.filename} {stat_res.st_dev} {stat_res.st_ino}) - attempting to claim lock")
        fcntl.lockf(self.file, fcntl.LOCK_EX)
        stat_res = os.stat(self.file.fileno())
        logging.debug(f"ExclusiveLockFile.lock({self.filename} {stat_res.st_dev} {stat_res.st_ino}) - locked")
        self.locked = True
        logging.debug("Locked {0}".format(self.filename))

    def unlock(self):
        if not self.locked:
            raise RuntimeError(f"Attempting to unlock {self.filename}, but I don't have that lock")
        logging.debug(f"ExclusiveLockFile.unlock({self.filename}) - releasing lock")
        fcntl.lockf(self.file, fcntl.LOCK_UN)
        self.locked = False
        logging.debug("Unlocked {0}".format(self.filename))

    def __enter__(self):
        self.lock()

    def __exit__(self, type, value, traceback):
        self.unlock()

    def __del__(self):
        if self.file is not None:
            self.file.close()

def samefile(f1, f2):
    """ Do file and/or path-like objects correspond to the same file on disk?

    Takes file-like objects (must provide .fileno() to get a file descriptor),
    file descriptors (ints), and path-like objects that can be passed to
    os.stat().

    All exceptions are eaten and the answer assumed to be False

    >>> from tempfile import mkdtemp
    >>> tmpdir = mkdtemp()
    >>> a = open(tmpdir+"/a","w")
    >>> b = open(tmpdir+"/b","w")
    >>> a2 = open(tmpdir+"/a","w")
    >>> samefile(a,b)
    False
    >>> samefile(a.fileno(),b)
    False
    >>> samefile(a,a2)
    True
    >>> import os
    >>> os.rename(tmpdir+"/a", tmpdir+"/c")
    >>> samefile(a,tmpdir+"/c")
    True
    >>> samefile(a,tmpdir+"/a")
    False
    >>> os.unlink(tmpdir+"/b")
    >>> os.unlink(tmpdir+"/c")
    >>> os.rmdir(tmpdir)
    """
    def desc_file(x):
        if x is None: return "None"
        if hasattr(x,'name'): return f'<{type(x).__name__} name="{x.name}">'
        return str(x)

    def dev_ino(file):
        if not isinstance(file, int) and hasattr(file,'fileno'):
            file = file.fileno()
        stat = os.stat(file)
        return (stat.st_dev, stat.st_ino)

    desc = f"samefile({desc_file(f1)}, {desc_file(f2)})"
    try:
        f1_s = dev_ino(f1)
        f2_s = dev_ino(f2)
        r = f1_s == f2_s
        logging.debug(f"{desc} (dev,ino) {f1_s} == {f2_s} -> {r}")
        return r
    except Exception as e:
        logging.debug(f"{desc} -> False by exception {e}")
        return False

def file_matches_filename(file, filename):
    """ Does file object file correspond to what is now at filename?

    This is a simple wrapper around samefile.

    WARNING: This adds an exclusive lock to filename!kkk

    All exceptions are eaten and the answer assumed to be False

    >>> from tempfile import mkdtemp
    >>> tmpdir = mkdtemp()
    >>> a = open(tmpdir+"/a","w")
    >>> b = open(tmpdir+"/b","w")
    >>> file_matches_filename(a,tmpdir+"/a")
    True
    >>> file_matches_filename(a,tmpdir+"/b")
    False
    >>> file_matches_filename(a,tmpdir+"/missingfile")
    False
    >>> import os
    >>> os.rename(tmpdir+"/a", tmpdir+"/c")
    >>> c = open(tmpdir+"/c","w")
    >>> newa = open(tmpdir+"/a","w")
    >>> file_matches_filename(a,tmpdir+"/a")
    False
    >>> file_matches_filename(a,tmpdir+"/c")
    True
    >>> import os
    >>> os.unlink(tmpdir+"/a")
    >>> os.unlink(tmpdir+"/b")
    >>> os.unlink(tmpdir+"/c")
    >>> os.rmdir(tmpdir)
    """
    logging.debug(f"Comparing file object ({file.name}) to {filename}")
    try:
        r = samefile(file.fileno(), filename)
        # Need a second lock because the .close()
        # will release one of them.
        logging.debug(f"  same? {r}")
        return r
    except FileNotFoundError as e:
        logging.debug(f"  Not the same, {filename} doesn't exist ({e})")
        return False
    except Exception as e:
        logging.warning(f"Trying to check if {filename} refers to an open file (which _might_ be namemd {file.name}. Ran into an unexpected exception: {e}")
        return False


def unlink_if_possible(filename):
    """ unlink a file; ignore errors

    >>> from csppfetch.exclusivelockfile import unlink_if_possible
    >>> open("/tmp/example-file-name", "w").close()
    >>> unlink_if_possible("/tmp/example-file-name")
    True
    >>> unlink_if_possible("/tmp/this-file-should-never-exist")
    False
    """
    try:
        os.unlink(filename)
        return True
    except:
        return False

class AtomicCreateIfMissing:
    """ Context manager to atomically create a file
    >>> from csppfetch.exclusivelockfile import AtomicCreateIfMissing
    >>> from tempfile import mkdtemp
    >>> tmpdir = mkdtemp()
    >>> filename = tmpdir+"/result"
    >>> #
    >>> with AtomicCreateIfMissing(filename,binary=False) as f:
    ...     if f is None:
    ...         raise Exception("This won't happen, as filename doesn't exist yet")
    ...     _ = f.write("this is the body of the test file")
    ...     try:
    ...         open(filename)
    ...         raise Exception("This won't happen; file doesn't exist yet")
    ...     except FileNotFoundError:
    ...         pass
    >>> # This file now exists
    >>> with open(filename) as f: print(f.read())
    this is the body of the test file
    >>> # Retrying will fail, returning None as the file
    >>> with AtomicCreateIfMissing(filename) as f:
    ...     if f is not None:
    ...         raise Exception("This won't happen, as the file already exists and AtomicCreateIfMissing returned None above")
    >>> import os
    >>> os.unlink(filename)
    >>> os.rmdir(tmpdir)
    """

    def __init__(self, filename, partial_suffix=".part", mkdirs=True, binary=True):
        self.filename = filename
        self.suffix = partial_suffix
        self.lockfilename = filename + partial_suffix
        self.lock = None
        self.file = None
        self.mkdirs = mkdirs
        self.binary = binary
        self.finished = False

    def __enter__(self):
        logging.debug(f"AtomicCreateIfMissing({self.filename} {self.lockfilename}).__enter__")
        if os.path.exists(self.filename):
            # While redundant, this short circuit succeeds in an expected
            # common case
            logging.debug(f"AtomicCreateIfMissing({self.filename}): File already exists; returning None")
            return None

        dir = os.path.dirname(self.filename)
        if len(dir) and not os.path.exists(dir):
            if self.mkdirs:
                logging.debug("Creating directories to {0}".format(dir))
                os.makedirs(dir)
            else:
                raise FileNotFoundError(f"{dir} does not exist, so I can't create {self.filename} in it. set mkdirs=True in AtomicCreateIfMissing to automatically try to create the directories")

        # append so we don't overwrite if someone else beat us
        # to it.
        mode = "a"
        if self.binary: mode += "b"
        f1 = open(self.lockfilename, mode)
        lock = ExclusiveLockFile(f1)
        lock.lock()
        logging.debug(f"AtomicCreateIfMissing claimed a lock on {self.lockfilename}")

        if file_matches_filename(f1, self.filename):
            logging.debug(f"AtomicCreateIfMissing({self.filename}): File created while I was trying to; returning None")
            lock.unlock()
            return None
        if not file_matches_filename(f1, self.lockfilename):
            lock.unlock()
            raise RuntimeError(f"Someting unexpected happened. I opened {self.lockfilename}, but it's not there anymore, and it didn't get moved to {self.filename}.")

        logging.debug(f"AtomicCreateIfMissing({self.filename}): I am handling creation")
        # Wipe anything some other poor process may have left behind
        # (but don't close() the file and lose our lock!)
        f1.seek(0)
        f1.truncate()

        self.lock = lock
        self.file = f1
        return f1

    def __exit__(self, type, value, traceback):
        if self.lock is not None:
            os.rename(self.lockfilename, self.filename)
            self.finished = True
            self._release()

    def _release(self):
        self.lock.unlock()
        self.lock = None
        self.file.close()
        self.file = None

    def abandon(self):
        if self.lock is None:
            return False
        self._release()
        return unlink_if_possible(self.lockfilename)

    def final_size(self):
        if self.finished:
            return os.path.getsize(self.filename)
        if self.file is not None:
            return self.file.tell()
        return 0



