#! /usr/bin/python3

import signal

# We need our own exception to ensure some other layer
# doesn't helpfully catch our timeout
class OutOfTime(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class TimeLimit:
    """ Limits how long code in a "with" block is allowed to run

    WARNING: Assumes its free to control SIGALRM. It will reset
    the SIGALRM handler when done, but it will stomp over any 
    existing alarm

    >>> import time
    >>> from csppfetch.timelimit import TimeLimit
    >>> with TimeLimit(1):
    ...     time.sleep(5)
    Traceback (most recent call last):
    ...
    csppfetch.timelimit.OutOfTime: TimeLimit expired
    >>> with TimeLimit(2):
    ...     time.sleep(0)
    ...     print("done")
    done
    """
    def __init__(self, max_time):
        """ Create TimeLimit object limited to max_time seconds
        """
        if (not isinstance(max_time,int)) or (max_time < 0):
            raise Exception(f"max_time argument to TimeLimit must be an positive, integer number of seconds (it's {max_time} of type {type(max_time)})")
        self.max_time = max_time
        self.timed_out = False
    def handle_alarm(self, signum, frame):
        self.timed_out = True
        raise OutOfTime("TimeLimit expired")
    def __enter__(self):
        signal.alarm(0)
        self.old_signal = signal.signal(signal.SIGALRM, self.handle_alarm)
        signal.alarm(self.max_time)
    def __exit__(self, type, value, traceback):
        signal.alarm(0)
        signal.signal(signal.SIGALRM, self.old_signal)

