#! /usr/bin/python3

import logging
import csppfetch
import csppfetch.exclusivelockfile
from csppfetch.exclusivelockfile import AtomicCreateIfMissing, ExclusiveLockFile

import os
import unittest
import urllib
import datetime
import tempfile
from tempfile import TemporaryDirectory
from datetime import timedelta



class ShadowEnvironment:
    """ Context manager to set environment variables, restoring previous when done

    >>> import os
    >>> os.environ["foo"] = "original value"
    >>> os.unsetenv("another")
    >>> with ShadowEnvironment(foo="bar", qux="zot") as se:
    ...     os.environ["foo"]
    ...     se.setenv("foo", "bar2") # Replaces previous
    ...     se.setenv("another","value")
    ...     os.environ["another"]
    ...     os.environ["foo"]
    'bar'
    'value'
    'bar2'
    >>> os.environ["foo"]
    'original value'
    >>> os.getenv("another", None) is None
    True
    """
    def __init__(self, **kwargs):
        self.newvars = kwargs
        self.oldvars = None
    def setenv(self, key, value):
        if value is not None:
            os.environ[key] = value
        else:
            os.unsetenv(key)
    def __enter__(self):
        self.oldvars = dict(os.environ)
        for key,value in self.newvars.items():
            self.setenv(key, value)
        return self
    def __exit__(self, type, value, traceback):
        os.environ.clear()
        os.environ.update(self.oldvars)
        self.oldvars = None


class DTestCase(unittest.TestCase):
    def assertFileExists(self, file):
        self.assertTrue(os.path.exists(file))
    def assertFileDoesNotExist(self, file):
        self.assertFalse(os.path.exists(file))
    def assertFileContentsAre(self, filename, expected):
        with open(filename) as f:
            actual = f.read()
        self.assertEqual(actual, expected)


        #with tempfile.NamedTemporaryFile() as f:

ELFT_MSG_WAITING = "waiting for lock"
ELFT_MSG_LOCKED = "claimed lock, waiting on pipe"
ELFT_MSG_UNLOCKED = "released lock"
ELFT_MSG_DONE = "exiting"

def ELFT_lock_proc(title, file, qresults, qcommands):
    qresults.put(f"{title} {ELFT_MSG_WAITING}")
    with ExclusiveLockFile(file):
        qresults.put(f"{title} {ELFT_MSG_LOCKED}")
        cmd = qcommands.get()
        if cmd != "unlock":
            msg = f"{title} received '{cmd} but expected 'unlock'"
            qresults.put(msg)
            raise Exception(msg)
    qresults.put(f"{title} {ELFT_MSG_UNLOCKED}")
    cmd = qcommands.get()
    if cmd != "done":
        qresults.put(f"{title} received {cmd} expected done")
        return 1
    qresults.put(f"{title} {ELFT_MSG_DONE}")

def ELFT_lock_proc_handle(title, file, qresults, qcommands):
    qresults.put(f"{title} {ELFT_MSG_WAITING}")
    with open(file, "a") as handle:
        with ExclusiveLockFile(handle):
            qresults.put(f"{title} {ELFT_MSG_LOCKED}")
            cmd = qcommands.get()
            if cmd != "unlock":
                msg = f"{title} received '{cmd} but expected 'unlock'"
                qresults.put(msg)
                raise Exception(msg)
        qresults.put(f"{title} {ELFT_MSG_UNLOCKED}")
        cmd = qcommands.get()
        if cmd != "done":
            qresults.put(f"{title} received {cmd} expected done")
            return 1
        qresults.put(f"{title} {ELFT_MSG_DONE}")
        
import multiprocessing
class ParallelTestProcess:
    def __init__(self, name, func, data):
        self.name = name
        self.data = data
        self.qresults = multiprocessing.Queue()
        self.qcommands = multiprocessing.Queue()
        self.process = multiprocessing.Process(target=func,
                args=[name, data, self.qresults, self.qcommands])
        self.running = False


    def start(self):
        if self.running:
            raise RuntimeError("trying to start() an already running process")
        self.process.start()
        logging.debug(f"ParallelTestProcess({self.name}).start() spawned {self.process.pid}")

        self.running = True

    def join(self, timeout=None):
        self.process.join(timeout)
        self.running = False

    def terminate(self):
        if self.running:
            self.process.terminate()
            self.running = False

    def is_alive(self):
        return self.process.is_alive()

    def get(self):
        m = self.qresults.get()
        logging.debug(f"ParallelTestProcess({self.name}).get() -> {m}")
        return m

    def put(self, command):
        logging.debug(f"ParallelTestProcess({self.name}).put({command})")
        return self.qcommands.put(command)

    def is_quiet(self):
        return self.qresults.empty()

    def __del__(self):
        logging.debug(f"ParallelTestProcess({self.name}).__del__()")
        self.terminate()

class ExclusiveLockFileTests(DTestCase):

    def impl_test_exclusivelockfile(self, proc):
        # Trying to deterministically test non-deterministic code is awful
        with TemporaryDirectory() as dir:
            p = []
            lockfile = dir+"/lock"
            open(lockfile,"x").close()

            lockfile = dir+"/lock"
            p = [ParallelTestProcess(f"p{i}", proc, lockfile) for i in range(3)]

            p[0].start()
            self.assertEqual(p[0].get(), f"p0 {ELFT_MSG_WAITING}")
            self.assertEqual(p[0].get(), f"p0 {ELFT_MSG_LOCKED}")
            self.assertTrue(p[0].is_quiet())
            p[1].start()
            self.assertEqual(p[1].get(), f"p1 {ELFT_MSG_WAITING}")
            self.assertTrue(p[0].is_quiet())
            self.assertTrue(p[1].is_quiet())
            p[0].put("unlock")
            self.assertEqual(p[0].get(), f"p0 {ELFT_MSG_UNLOCKED}")
            self.assertTrue(p[0].is_quiet())
            self.assertEqual(p[1].get(), f"p1 {ELFT_MSG_LOCKED}")
            self.assertTrue(p[1].is_quiet())
            p[0].put("done")
            self.assertEqual(p[0].get(), f"p0 {ELFT_MSG_DONE}")
            self.assertTrue(p[0].is_quiet())
            self.assertTrue(p[1].is_quiet())
            p[1].put("unlock")
            self.assertEqual(p[1].get(), f"p1 {ELFT_MSG_UNLOCKED}")
            self.assertTrue(p[1].is_quiet())
            p[1].put("done")
            self.assertEqual(p[1].get(), f"p1 {ELFT_MSG_DONE}")
            self.assertTrue(p[1].is_quiet())

            # The other locks are gone, this should be trivial
            p[2].start()
            self.assertEqual(p[2].get(), f"p2 {ELFT_MSG_WAITING}")
            self.assertEqual(p[2].get(), f"p2 {ELFT_MSG_LOCKED}")
            self.assertTrue(p[2].is_quiet())
            p[2].put("unlock")
            self.assertEqual(p[2].get(), f"p2 {ELFT_MSG_UNLOCKED}")
            self.assertTrue(p[2].is_quiet())
            p[2].put("done")
            self.assertEqual(p[2].get(), f"p2 {ELFT_MSG_DONE}")
            self.assertTrue(p[2].is_quiet())

            for i in range(3):
                MAXIMUM_TIME_TO_EXIT=2 # seconds
                p[i].join(MAXIMUM_TIME_TO_EXIT)
                self.assertFalse(p[i].is_alive())

    def test_exclusivelockfile_name(self):
        self.impl_test_exclusivelockfile(ELFT_lock_proc)

    def test_exclusivelockfile_handle(self):
        self.impl_test_exclusivelockfile(ELFT_lock_proc_handle)


ACIM_MSG_WAITING = "waiting for lock"
ACIM_MSG_LOCKED = "claimed lock, waiting on pipe"
ACIM_MSG_TOO_LATE = "file already exists"
ACIM_MSG_UNLOCKED = "released lock"
ACIM_MSG_DONE = "exiting"
ACIM_FILE_CONTENTS_FMT = "AtomicCreateIfMissing test file {}"

def ACIM_proc(title, file, qresults, qcommands):
    logging.debug(f"ACIM_proc {title} start")
    qresults.put(f"{title} {ACIM_MSG_WAITING}")
    with AtomicCreateIfMissing(file, binary=False) as f:
        if f is None:
            logging.debug(f"ACIM_proc {title} {ACIM_MSG_TOO_LATE}")
            qresults.put(f"{title} {ACIM_MSG_TOO_LATE}")
            return
        logging.debug(f"ACIM_proc {title} {ACIM_MSG_LOCKED}")
        qresults.put(f"{title} {ACIM_MSG_LOCKED}")
        cmd = qcommands.get()
        if cmd != "write":
            msg = f"{title} received '{cmd} but expected 'write'"
            logging.error(msg)
            qresults.put(msg)
            raise Exception(msg)
        f.write(ACIM_FILE_CONTENTS_FMT.format(title))
        logging.debug(f"ACIM_proc {title} written")
    qresults.put(f"{title} {ACIM_MSG_UNLOCKED}")
    cmd = qcommands.get()
    if cmd != "done":
        qresults.put(f"{title} received {cmd} but expected 'done'")
        logging.error(msg)
        return 1
    logging.debug(f"ACIM_proc {title} finished")
    qresults.put(f"{title} {ACIM_MSG_DONE}")


class AtomicCreateIfMissingTests(DTestCase):

    def test_basics(self):
        from csppfetch.exclusivelockfile import AtomicCreateIfMissing
        with TemporaryDirectory() as tmpdir:
            filename = tmpdir+"/final"

            with AtomicCreateIfMissing(filename,binary=False) as f:
                self.assertIsNotNone(f)
                f.write("this is the body of the test file")
                self.assertFileDoesNotExist(filename)

            self.assertFileExists(filename)
            self.assertFileContentsAre(filename,
                    "this is the body of the test file")

            with AtomicCreateIfMissing(filename) as f:
                self.assertIsNone(f)


    def test_parallel(self):
        with TemporaryDirectory() as dir:
            lockfile = dir+"/new-file"
            p = [ParallelTestProcess(f"p{i}", ACIM_proc, lockfile) for i in range(3)]

            import time
            p[0].start()
            self.assertEqual(p[0].get(), f"p0 {ACIM_MSG_WAITING}")
            self.assertEqual(p[0].get(), f"p0 {ACIM_MSG_LOCKED}")
            self.assertTrue(p[0].is_quiet())
            p[1].start()
            self.assertEqual(p[1].get(), f"p1 {ACIM_MSG_WAITING}")
            self.assertTrue(p[0].is_quiet())
            self.assertTrue(p[1].is_quiet())
            p[0].put("write")
            self.assertEqual(p[0].get(), f"p0 {ACIM_MSG_UNLOCKED}")
            self.assertTrue(p[0].is_quiet())
            self.assertEqual(p[1].get(), f"p1 {ACIM_MSG_TOO_LATE}")
            p[1].join(1) # A moment to get its affairs in order
            self.assertFalse(p[1].is_alive())
            self.assertTrue(p[1].is_quiet())
            p[0].put("done")
            self.assertEqual(p[0].get(), f"p0 {ACIM_MSG_DONE}")
            p[0].join(1) # A moment to get its affairs in order
            self.assertFalse(p[0].is_alive())
            self.assertTrue(p[0].is_quiet())

            # The other locks are gone, this should be trivial
            p[2].start()
            self.assertEqual(p[2].get(), f"p2 {ACIM_MSG_WAITING}")
            self.assertEqual(p[2].get(), f"p2 {ACIM_MSG_TOO_LATE}")
            p[2].join(1) # A moment to get its affairs in order
            self.assertFalse(p[2].is_alive())
            self.assertTrue(p[2].is_quiet())



class DownloadFunctionsTests(DTestCase):

    def test_download_already_present_writable(self):
        with TemporaryDirectory() as dirpath:
            filepath = dirpath+"/test"
            f = open(filepath,"wb")
            f.write(b"x"*100)
            f.close()
            stats = csppfetch.DownloadStatistics()
            csppfetch.download_already_present("Example Download", f.name, stats, True)
            csppfetch.download_already_present("Example Download", f.name, stats, True)
            self.assertEqual(stats.cache_hits, 2)
            self.assertEqual(stats.cache_size, 100*2)

    def test_download_already_present_non_writable(self):
        with TemporaryDirectory() as dirpath:
            filepath = dirpath+"/test"
            open(filepath,"w").close()
            os.chmod(dirpath, 0)
            csppfetch.download_already_present("Example Download", filepath, None, False)
            with self.assertLogs(level='WARNING') as logs:
                with self.assertRaises(Exception):
                    try:
                        csppfetch.download_already_present("Example Download", filepath, None, True)
                    except:
                        raise
                    finally:
                        os.chmod(dirpath, 0o700)
            self.assertRegex(logs.output[0], r'^ERROR:root:Unable to update timestamp')

    def test_download_successful(self):
        download_stats = csppfetch.DownloadStatistics()
        with self.assertLogs(level='DEBUG') as logs:
            with TemporaryDirectory() as dirpath:
                dst = dirpath+"/test"
                f = open(dst,"w")
                f.write("X"*123)
                f.close()
                csppfetch.download_successful("Test case", dst, os.path.getsize(dst), download_stats)
        self.assertEqual(download_stats.downloaded_files, 1)
        self.assertEqual(download_stats.downloaded_size, 123)
        self.assertRegex(logs.output[0],r"^DEBUG:root:.* downloaded \(Test case\).")

    def test_download_to_file_no_lock_success(self):
        with tempfile.NamedTemporaryFile() as f:
            name = f.name
            csppfetch.download_to_file_no_lock("http://www.google.com/", "fake_path", f)
            self.assertFileExists(name)
            self.assertGreater(os.path.getsize(name), 0)

    def test_download_to_file_no_lock_failure(self):
        with TemporaryDirectory() as dir:
            dst = dir+"/test"
            stats = csppfetch.DownloadStatistics()
            with self.assertRaises(urllib.error.HTTPError):
                csppfetch.download_to_file_no_lock("http://www.google.com/should-never-exist-404-please", dst, stats)
            self.assertFileDoesNotExist(dst)
            self.assertEqual(stats.downloaded_files,0)
            self.assertEqual(stats.downloaded_size,0)
            self.assertEqual(stats.cache_hits,0)
            self.assertEqual(stats.cache_size,0)

    def test_download_all_success(self):
        with TemporaryDirectory() as dir:
            stats = csppfetch.DownloadStatistics()
            url_to_dst = {
                 "http://www.google.com/": dir+"/a",
                 "http://www.microsoft.com/": dir+"/b",
                 }
            csppfetch.download_all(url_to_dst, stats, True, 60*5)
            self.assertFileExists(dir+"/a")
            self.assertGreater(os.path.getsize(dir+"/a"), 0)
            self.assertFileExists(dir+"/b")
            self.assertGreater(os.path.getsize(dir+"/b"), 0)
            self.assertEqual(stats.downloaded_files,2)
            self.assertGreater(stats.downloaded_size,0)
            self.assertEqual(stats.cache_hits,0)
            self.assertEqual(stats.cache_size,0)

    def test_download_all_timeout(self):
        with TemporaryDirectory() as dir:
            stats = csppfetch.DownloadStatistics()
            url_to_dst = { f"http://geodb2.ssec.wisc.edu/ancillary/2019_05_27_147/gfs.19052700_F{f:03d}.hdf" : dir+"/f" for f in range(3,22,3) }

            with self.assertRaises(csppfetch.timelimit.OutOfTime):
                csppfetch.download_all(url_to_dst, stats, True, 1)
            for file in url_to_dst.values():
                self.assertFileDoesNotExist(file)

    def test_download_all_partial_existance(self):
        with TemporaryDirectory() as dir:
            with open(dir+"/a","w") as f: f.write("XXX")
            stats = csppfetch.DownloadStatistics()
            url_to_dst = {
                 "http://www.google.com/": dir+"/a",
                 "http://www.microsoft.com/": dir+"/b",
                 }
            csppfetch.download_all(url_to_dst, stats, True, 60*5)
            self.assertFileExists(dir+"/a")
            self.assertEqual(os.path.getsize(dir+"/a"), 3)
            self.assertFileExists(dir+"/b")
            self.assertGreater(os.path.getsize(dir+"/b"), 0)
            self.assertEqual(stats.downloaded_files,1)
            self.assertGreater(stats.downloaded_size,0)
            self.assertEqual(stats.cache_hits,1)
            self.assertEqual(stats.cache_size,3)

    def test_download_all_partial_failed(self):
        # test if one file fails to download/be present
        with TemporaryDirectory() as dir:
            stats = csppfetch.DownloadStatistics()
            url_to_dst = {
                 "http://www.google.com/": dir+"/a",
                 "http://www.google.com/should-never-exist-404-please": dir+"/b",
                 }
            with self.assertRaises(urllib.error.HTTPError):
                csppfetch.download_all(url_to_dst, stats, True, 60*5)

    def test_download_all_no_download_all_present(self):
        # test if downloads disabled and files exist
        with TemporaryDirectory() as dir:
            for name in ("/a","/b"):
                with open(dir+name,"w") as f:
                    f.write("XXX")
            stats = csppfetch.DownloadStatistics()
            url_to_dst = {
                 "http://www.google.com/": dir+"/a",
                 "http://www.microsoft.com/": dir+"/b",
                 }
            csppfetch.download_all(url_to_dst, stats, False, 60*5)
            self.assertFileExists(dir+"/a")
            self.assertEqual(os.path.getsize(dir+"/a"), 3)
            self.assertFileExists(dir+"/b")
            self.assertEqual(os.path.getsize(dir+"/b"), 3)
            self.assertEqual(stats.downloaded_files,0)
            self.assertEqual(stats.downloaded_size,0)
            self.assertEqual(stats.cache_hits,2)
            self.assertEqual(stats.cache_size,6)

    def test_download_all_no_download_missing(self):
        # test if downloads disabled and a file is missing
        with TemporaryDirectory() as dir:
            with open(dir+"/a","w") as f: f.write("XXX")
            stats = csppfetch.DownloadStatistics()
            url_to_dst = {
                 "http://www.google.com/": dir+"/a",
                 "http://www.microsoft.com/": dir+"/b",
                 }
            with self.assertRaises(Exception):
                csppfetch.download_all(url_to_dst, stats, False, 60*5)

class FileSetTests(unittest.TestCase):
    SOME_FILES = {
            "some/path/to/final-file-a.nc": "output-a.nc",
            "some/path/to/final-file-b.nc": "output-b.nc"
        }
    def test_basic(self):
        a = csppfetch.FileSet( dict(self.SOME_FILES), False, 'description of fileset here')
        self.assertEqual(a.expected, False)
        self.assertEqual(a.description, 'description of fileset here')
        self.assertEqual(a.urls_to_files, self.SOME_FILES)

    def test_equality(self):
        a = csppfetch.FileSet( dict(self.SOME_FILES), False, 'description of fileset here')

        b = csppfetch.FileSet( dict(self.SOME_FILES), False, 'description of fileset here')
        self.assertEqual(a,b)

        b = csppfetch.FileSet( dict(self.SOME_FILES), True, 'description of fileset here')
        self.assertNotEqual(a,b)
        b = csppfetch.FileSet( dict(self.SOME_FILES), False, 'different description')
        self.assertNotEqual(a,b)
        newdict = dict(self.SOME_FILES)
        newdict["a"] = "b"
        b = csppfetch.FileSet( newdict , False, 'description of fileset here')
        self.assertNotEqual(a,b)

        

class DownloaderTests(DTestCase):

    # The test Downloader is initialized
    # with these values, which are also useful
    # to tests when building assertions
    PRIORITIES = ['', '_preliminary']
    OLDEST_USABLE = timedelta(days=3)
    OLDEST_MIRROR = timedelta(days=3)
    PERIOD = timedelta(days=1)
    PACKAGE_ENV_ID = "DOWNLOADER_TEST_"
    URL_BASE = "https://geodb.ssec.wisc.edu/ancillary/"
    NAME = 'Example Downloader for Testing'

    def safe_test_date(self):
        """ Return datetime.datetime for date where geodb.ssec.wisc.edu keeps data forever """
        return datetime.datetime(2019,5,27,12,0,0)


    def make_test_downloader(self, oldest_usable=OLDEST_USABLE, oldest_mirror=OLDEST_MIRROR):
        sst_filename = "avhrr-only-v2.%Y%m%d{priority}.nc"
        period = timedelta(days=1)
        sst = csppfetch.Downloader(
            name = self.NAME,
            package_env_id = self.PACKAGE_ENV_ID,
            url_base = self.URL_BASE,
            url_relative = "%Y_%m_%d_%j/"+sst_filename,
            local = sst_filename,
            period = self.PERIOD,
            priorities = self.PRIORITIES,
            oldest_usable = oldest_usable,
            oldest_mirror = oldest_mirror,
            expected_newest = timedelta(hours=1)
            )
        return sst

    def test_get_base_urls(self):
        d = self.make_test_downloader()
        self.assertEqual(d.get_base_urls(), [self.URL_BASE])

    def test_priorities(self):
        d = self.make_test_downloader()
        test_datetime = datetime.datetime(2020,6,1,12,0,0)
        self.assertEqual(d._priorities(test_datetime), self.PRIORITIES)

    def test_nearest_preceeding_time(self):
        d = self.make_test_downloader()
        test_datetime = datetime.datetime(2020,6,1,12,0,0)
        prev_time = d._nearest_preceeding_time(test_datetime)
        self.assertEqual(prev_time, datetime.datetime(2020,6,1,0,0,0))

    def test_previous_time(self):
        d = self.make_test_downloader()
        test_datetime = datetime.datetime(2020,6,1,12,0,0)
        prev_time = d._previous_time(test_datetime)
        self.assertEqual(prev_time, datetime.datetime(2020,6,1,0,0,0))
        self.assertEqual(d._previous_time(prev_time), datetime.datetime(2020,5,31,0,0,0))

    def test_get_filesets_for_time(self):
        d = self.make_test_downloader()
        test_datetime = datetime.datetime(2020,6,1,12,0,0)

        # This can be shorter than the full result; it should contain a first
        # entries.
        expected_filesets = [
            csppfetch.FileSet( {"2020_06_01_153/avhrr-only-v2.20200601.nc": "avhrr-only-v2.20200601.nc"}, True, self.NAME+' for 2020-06-01 00:00:00 at priority "" (empty string)'),
            csppfetch.FileSet( {"2020_06_01_153/avhrr-only-v2.20200601_preliminary.nc": "avhrr-only-v2.20200601_preliminary.nc"}, True, self.NAME+' for 2020-06-01 00:00:00 at priority "_preliminary"'),
            csppfetch.FileSet( {"2020_05_31_152/avhrr-only-v2.20200531.nc": "avhrr-only-v2.20200531.nc"}, True, self.NAME+' for 2020-05-31 00:00:00 at priority "" (empty string)'),
                ]

        all_found_filesets = d.get_filesets_for_time(test_datetime)

        num_found_filesets = len(all_found_filesets)
        num_expected_filesets = len(self.PRIORITIES) * (self.OLDEST_USABLE/self.PERIOD)
        self.assertEqual(num_found_filesets, num_expected_filesets)

        found_filesets = all_found_filesets[0:len(expected_filesets)]
        self.assertEqual(found_filesets, expected_filesets)

    def test_download_for_time(self):
        d = self.make_test_downloader()
        test_datetime = self.safe_test_date()
        outfile = "avhrr-only-v2.20190527.nc"
        with TemporaryDirectory() as dirpath:
            d.download_for_time(test_datetime, dirpath) # do_download = True, download_stats = None
            self.assertFileExists(dirpath+"/"+outfile)

    def test_download_for_time_404(self):
        # Override days to speed up test
        d = self.make_test_downloader(oldest_usable=datetime.timedelta(days=1))
        # This data set should be on geodb.ssec.wisc.edu indefinitely
        test_datetime = datetime.datetime(1999,1,1,12,0,0)
        outfile = "avhrr-only-v2.19990101.nc"
        with TemporaryDirectory() as dirpath:
            with self.assertLogs(level='ERROR') as logs:
                d.download_for_time(test_datetime, dirpath,retries=1) # do_download = True, download_stats = None
                self.assertFileDoesNotExist(dirpath+"/"+outfile)
        self.assertRegex(logs.output[0], r'^ERROR:root:Unable to download any file set')
        self.assertRegex(logs.output[1], r'HTTP Error 404')

    def test_get_mirror_filesets(self):
        d = self.make_test_downloader()
        test_datetime = self.safe_test_date()
        timesteps = d.get_mirror_filesets(end=test_datetime)
        # This is a hard coded sanity test before getting to the 
        # overly clever dynamically generated test
        self.assertEqual(repr(timesteps[0][0]),'<Example Downloader for Testing for 2019-05-27 00:00:00 at priority "" (empty string) (expected? True): {2019_05_27_147/avhrr-only-v2.20190527.nc: avhrr-only-v2.20190527.nc}>')
        for timestep in range(3):
            idx = 0
            days = range(27,22,-1)
            subdays = days[timestep:timestep+4]
            for day in subdays:
                day_of_year = day+120
                for prio,desc in {'': '"" (empty string)', '_preliminary': '"_preliminary"'}.items():
                    expected = f'<Example Downloader for Testing for 2019-05-{day} 00:00:00 at priority {desc} (expected? True): {{2019_05_{day}_{day_of_year}/avhrr-only-v2.201905{day}{prio}.nc: avhrr-only-v2.201905{day}{prio}.nc}}>'
                    self.assertEqual(repr(timesteps[timestep][idx]),expected)
                    idx += 1



    def test_mirror_simple(self):
        d = self.make_test_downloader(oldest_mirror=timedelta(days=1))
        test_datetime = self.safe_test_date()
        # Tomorrow can use today's data
        end = test_datetime + timedelta(days=1)
        with TemporaryDirectory() as dirpath:
            d.mirror(dirpath, end_time=end)
            self.assertFileExists(dirpath+"/avhrr-only-v2.20190527.nc")

    def mock_file_contents(self, path):
        basename = os.path.basename(path)
        return f"I am a test file named {basename}"

    def mock_file(self, path, contents = None):
        """ create file "path", making directories if needed """
        os.makedirs(os.path.dirname(path),exist_ok=True)
        if contents is None:
            contents = self.mock_file_contents(path)
        with open(path, "w") as f:
            f.write(contents)

    def assertMockFile(self, path, expected = None):
        failed = None
        try:
            with open(path) as f:
                actual = f.read()
        except Exception as e:
            failed = f"Unable to read {path}: {e}"
        if failed is not None:
            self.fail(failed)
        if expected is None:
            expected = self.mock_file_contents(path)
        if actual != expected:
            self.fail(f"expected of {path} are {actual} instead of expected {expected}")


    def test_mirror_complex(self):
        first_date = datetime.datetime(2017,1,1)
        last_date = first_date + timedelta(days=4)
        d = self.make_test_downloader(oldest_mirror=timedelta(days=4))

        with TemporaryDirectory() as mirror:

            self.mock_file(f"{mirror}/2017_01_01_001/avhrr-only-v2.20170101_preliminary.nc")
            self.mock_file(f"{mirror}/2017_01_02_002/avhrr-only-v2.20170102.nc")
            self.mock_file(f"{mirror}/2017_01_03_003/avhrr-only-v2.20170103.nc")
            self.mock_file(f"{mirror}/2017_01_04_004/avhrr-only-v2.20170104.nc")
            self.mock_file(f"{mirror}/2017_01_05_005/avhrr-only-v2.20170105.nc")

            with ShadowEnvironment(DOWNLOADER_TEST_URL=f"file://{mirror}/"),\
                    TemporaryDirectory() as outdir:

                d.mirror(outdir, end_time=last_date)
                self.assertMockFile(outdir+"/avhrr-only-v2.20170101_preliminary.nc")
                self.assertMockFile(outdir+"/avhrr-only-v2.20170102.nc")
                self.assertMockFile(outdir+"/avhrr-only-v2.20170103.nc")
                self.assertMockFile(outdir+"/avhrr-only-v2.20170104.nc")
                self.assertMockFile(outdir+"/avhrr-only-v2.20170105.nc")


def add_module_doctest(tests, module, context):
    import warnings
    import doctest
    def suppress(module):
        for mod in (module.__name__, 'doctest'):
            warnings.filterwarnings("ignore",
                    message= r'unclosed file <_io.TextIOWrapper',
                    category=ResourceWarning,
                    module=mod+"$")
    tests_before = tests.countTestCases()
    tests.addTests(doctest.DocTestSuite(module, setUp=lambda x:suppress(module), globs=context))
    tests_after = tests.countTestCases()
    if tests_before >= tests_after:
        raise Exception(f"Had {tests_before}, added doctests for {module.__name__}, then had {tests_after}, which is not more. Seems bad.")


def load_tests(loader, tests, ignore):
    import doctest

    from datetime import timedelta
    import datetime
    sst_filename = "avhrr-only-v2.%Y%m%d{priority}.nc"
    sst = csppfetch.Downloader(
        name = "Example Downloader",
        package_env_id = 'EXAMPLE_',
        url_base = "https://geodb.ssec.wisc.edu/ancillary/",
        url_relative = "%Y_%m_%d_%j/"+sst_filename,
        local = sst_filename,
        period = timedelta(hours=12),
        epoch_start = datetime.datetime(2010,1,1,3,0,0),
        priorities = ['final', 'preliminary'],
        oldest_usable = timedelta(days=1),
        oldest_mirror = timedelta(days=2),
        expected_newest = timedelta(hours=1)
        )


    # We need this so we can have "d" for Downloader examples
    # We could just use globals() as a base, but it would
    # include things we've imported, which would be misleading
    # in documentation.
    context = {
        # Objects
        'Downloader': csppfetch.Downloader,
        'FileSet': csppfetch.FileSet,
        'ShadowEnvironment': ShadowEnvironment,

        # Functions
        'download_all': csppfetch.download_all,
        'download_already_present': csppfetch.download_already_present,
        'download_to_file_no_lock': csppfetch.download_to_file_no_lock,
        'samefile': csppfetch.exclusivelockfile.samefile,
        'file_matches_filename': csppfetch.exclusivelockfile.file_matches_filename,

        # Examples in documentation
        'd': sst,
        }

    tests.addTests(doctest.DocTestSuite(globs=context))

    add_module_doctest(tests, csppfetch, context)

    import csppfetch.exclusivelockfile as exclusivelockfile
    add_module_doctest(tests, exclusivelockfile, context)

    import csppfetch.daterange as daterange
    add_module_doctest(tests, daterange, context)

    add_module_doctest(tests, csppfetch.timelimit, context)

    return tests

if __name__ == '__main__':
    import logging
    #logging.basicConfig(format='%(process)d:%(levelname)s:%(message)s', level=logging.DEBUG)

    unittest.main()
